# -*- coding: utf-8 -*-

"""
Package et-micc2
================

Top-level package for :py:mod:`et-micc2`.
"""
__version__ = "2.5.2"

import et_micc2.cli_micc # avoid attribute error when executing micc when installed from pypi
